# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from types import NotImplementedType

from .basemodel import BaseModel
from .modelidentity import ModelIdentity
from .modelautoassignedidentity import ModelAutoAssignedIdentity


class PersistedModel(BaseModel):
    __abstract__: bool = True
    __surrogate__: ModelAutoAssignedIdentity | NotImplementedType | None | int
    __key__: ModelIdentity | NotImplementedType

    def natural_key(self) -> tuple[str, ...]:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import Awaitable
from typing import Callable
from typing import TypeVar

from ckms.core import parse_spec
from ckms.core import Keychain
from ckms.utils import b64encode_str
from ckms.utils import b64decode

from cbra.core.conf import settings
from cbra.types import IDependant

T = TypeVar('T', bound='SecretKey')
SIGNING_KEY_NAME: str = 'sig'


class SecretKey(IDependant):
    """The key used by the application to perform various signing operations,
    such as signing a session cookie. Only one instance per application should
    be created, since it used a single instance of its key.
    """
    __module__: str = 'cbra.core.params'
    signing_key_name: str = SIGNING_KEY_NAME
    keychain: Keychain = Keychain()

    @property
    def signing_key(self):
        assert self.keychain.has(self.signing_key_name) # nosec
        return self.keychain.get(self.signing_key_name)

    @classmethod
    async def setup(cls: type[T]) -> T:
        secret_key = settings.SECRET_KEY
        if secret_key and re.match(r'^(local|azure|google)\:', secret_key):
            raise NotImplementedError
        self = cls()
        if secret_key and not cls.keychain.has(cls.signing_key_name):
            spec = parse_spec({
                'provider': 'local',
                'algorithm': 'HS384',
                'kty': 'oct',
                'use': 'sig',

                # TODO: The parameter is incorrectly named but will do for now since
                # there is no public documentation.
                'key': {'cek': secret_key}
            })
            await cls.keychain.add(cls.signing_key_name, spec)
        return self

    def __init__(self):
        self.keychain = SecretKey.keychain

    async def sign(self, value: bytes | str, encoding: str = 'utf-8') -> str:
        """Sign the given value."""
        if isinstance(value, str):
            value = str.encode(value, encoding=encoding)
        return b64encode_str(await self.signing_key.sign(value))

    async def verify(self, value: bytes | str, message: bytes) -> bool:
        if isinstance(value, str):
            value = b64decode(value)
        return await self.signing_key.verify(value, message)

    @classmethod
    def __inject__(cls: type[T]) -> Callable[..., Awaitable[T] | T]:
        return cls.setup


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Callable
from typing import TypeVar

from pydantic import Field

from .application import Application
from .endpoint import Endpoint
from . import ioc
from .params import *
from .resource import Create
from .resource import Delete
from .resource import Mutable
from .resource import Replace
from .resource import Resource
from .resource import ResourceModel
from .resource import ResourceType
from .resource import Retrieve
from .resource import Update
from .secretkey import SecretKey


T = TypeVar('T')


__all__: list[str] = [
    'describe',
    'inject',
    'instance',
    'ioc',
    'permission',
    'Application',
    'ApplicationSecretKey',
    'Create',
    'Delete',
    'Endpoint',
    'Field',
    'Mutable',
    'Replace',
    'Resource',
    'ResourceModel',
    'ResourceType',
    'Retrieve',
    'SecretKey',
    'Update'
]

inject = ioc.inject
instance = ioc.instance
permission = Endpoint.require_permission


class describe:
    status_code: int = 200
    summary: str | None = None

    def __init__(
        self,
        status_code: int = 200,
        summary: str | None = None
    ) -> None:
        self.status_code = status_code

    def __call__(
        self,
        func: Callable[..., T]
    ) -> Callable[..., T]:
        if not hasattr(func, 'params'):
            func.params = {} # type: ignore
        func.params.update({ # type: ignore
            k: v for k, v in self.__dict__.items()
            if v is not None
        })
        return func


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

import pydantic

from .detailaction import DetailAction


T = TypeVar('T', bound='UpdateAction')


class UpdateAction(DetailAction):
    action: str = 'update'
    name_template: str = 'Update {article} {name}'

    @classmethod
    def fromfunc(
        cls: type[T],
        name: str,
        func: Any
    ) -> T:
        return cls(name=name, method='PATCH', func=func)

    def can_write(self) -> bool:
        return True

    def get_return_annotation(self) -> Any:
        return self.endpoint.model.__response_model__

    def get_write_model(self) -> type[pydantic.BaseModel]:
        return self.endpoint.model.__update_model__


# Copyright (C) 2020-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`PubsubMessage`."""
import datetime
import binascii
import base64
import json
from typing import Any

import pydantic


class PubsubMessage(pydantic.BaseModel):
    """A message that is published by publishers and consumed by subscribers.
    The message must contain either a non-empty `data` field or at least one
    attribute (in the `attributes` object). Note that client libraries represent
    this object differently depending on the language.
    """
    message_id: str = pydantic.Field(
        default=...,
        alias='messageId',
        title="Message ID",
        description=(
            "ID of this message, assigned by the server when the message is "
            "published. Guaranteed to be unique within the topic. This value "
            "may be read by a subscriber that receives a `PubsubMessage` via "
            "a `subscriptions.pull` call or a push delivery. It must not be "
            "populated by the publisher in a `topics.publish` call."
        )
    )

    publish_time: datetime.datetime = pydantic.Field(
        default=...,
        alias='publishTime',
        title="Published",
        description=(
            "The time at which the message was published, populated by Google "
            "Pub/Sub when it receives the `topics.publish` call. It must not "
            "be populated by the publisher in a `topics.publish` call.\n\nA "
            "timestamp in RFC3339 UTC \"Zulu\" format, with nanosecond "
            "resolution and up to nine fractional digits. Examples: `2014-10-"
            "02T15:01:23Z` and `2014-10-02T15:01:23.045123456Z`."
        )
    )
    attributes: dict[str, Any] | None = pydantic.Field(
        default=None,
        title="Attributes",
        description=(
            "Attributes for this message. If this field is empty, the "
            "message contains non-empty data. This can be used to filter "
            "messages on the subscription.\n\n"
            "An object containing a list of `key`: `value` pairs. Example: "
            "`{ \"name\": \"wrench\", \"mass\": \"1.3kg\", \"count\": \"3\" }`."
        )
    )

    data: str | None = pydantic.Field(
        default=None,
        title="Data",
        description=(
            "The message data field. If this field is empty, the message "
            "contains at least one attribute.\n\nA base64-encoded string."
        )
    )

    def get_data(self) -> dict[str, Any] | list[Any] | None:
        """Return a dictionary or a list containing the message
        data as specified by the ``.data`` attribute. The encoding
        is assumed JSON/UTF-8.
        """
        if self.data is None: # pragma: no cover
            return None
        try:
            return json.loads(base64.b64decode(self.data))
        except (binascii.Error, json.decoder.JSONDecodeError, TypeError):
            raise ValueError



# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import logging
from datetime import datetime
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Literal
from typing import TypeVar

import fastapi

from .abortable import Abortable
from .iauthorizationcontextfactory import IAuthorizationContextFactory
from .irequestprincipal import IRequestPrincipal
from .iroutable import IRoutable
from .isessionmanager import ISessionManager
from .forbidden import Forbidden
from .notauthorized import NotAuthorized


T = TypeVar('T', bound='IEndpoint')


class IEndpoint:
    __module__: str = 'cbra.types'
    allowed_http_methods: list[str]
    context_factory: IAuthorizationContextFactory
    handlers: list[IRoutable]
    include_in_schema: bool = True
    name: str | None = None
    logger: logging.Logger = logging.getLogger('uvicorn')
    session: ISessionManager[Any]
    with_options: bool = True

    #: The set of permissions supported by this endpoint. These must be
    #: defined beforehand to limit the number of calls to remote IAM
    #: systems.
    permissions: set[str]

    #: The list of subjects that may access this endpoint. Override this
    #: property if an endpoint needs to hardcode the allowed subjects
    #: that may invoke it, for example when expecting a request from
    #: Google using a priorly known service account.
    allowed_subjects: set[str] = set()

    #: The set of trusted authorization servers. This will override the
    #: :attr:`cbra.core.conf.settings.TRUSTED_AUTHORIZATION_SERVERS`
    #: setting.
    trusted_providers: set[str] = set()

    principal: IRequestPrincipal
    request: fastapi.Request
    response: fastapi.Response
    router: fastapi.APIRouter

    #: Indicates if all requests to the endpoint must be authenticated.
    require_authentication: bool = False
    status_code: int = 200
    summary: str | None = None
    tags: list[str] = []

    @staticmethod
    def require_permission(name: str) -> Callable[..., Any]:
        """Decorate a method on an :class:`~cbra.types.IEndpoint` implementation
        to require the given permission. If the request does not have permission
        `name`, then :class:`~cbra.types.Forbidden` is raised.
        """
        def decorator_factory(
            func: Callable[..., Any]
        ) -> Callable[['IEndpoint'], Awaitable[Any]]:
            @functools.wraps(func)
            async def f(self: 'IEndpoint', *args: Any, **kwargs: Any) -> Any:
                if not await self.is_authorized(name):
                    raise Forbidden
                return await func(self, *args, **kwargs)
            return f
        return decorator_factory

    @classmethod
    def configure(cls: type[T], overrides: dict[str, Any]) -> type[T]:
        return type(cls.__name__, (cls,), kwargs) # type: ignore

    async def get_allowed_subjects(self) -> set[str]:
        """Return the set of allowed subjects that may access this resource.
        By default, returns :attr:`allowed_subjects`.
        """
        return self.allowed_subjects

    def get_success_headers(self, data: Any) -> dict[str, Any]:
        """Return a mapping holding the headers to add on a successful
        request based on the return value of the request handler.
        """
        return {}

    async def run_handler(
        self,
        func: Callable[..., Any],
        *args: Any,
        **kwargs: Any
    ):
        try:
            await self.authenticate()
            if self.require_authentication and not self.ctx.is_authenticated():
                raise NotAuthorized
            return await func(self, *args, **kwargs)
        except Abortable as exc:
            return await exc.as_response()

    async def authenticate(self) -> None:
        self.ctx = await self.context_factory.authenticate(
            request=self.request,
            principal=self.principal,
            providers=self.trusted_providers,
            subjects=await self.get_allowed_subjects()
        )

    async def is_authorized(self, name: str) -> bool:
        """Return a boolean if the given authorization context has a
        certain permission.
        """
        await self.ctx.authorize()
        return self.has_permission(name)

    def has_permission(self, name: str) -> bool:
        """Return a boolean if the request has the given permission."""
        return self.ctx.has_permission(name)

    def set_cookie(
        self,
        key: str,
        value: str = "",
        max_age: int | None = None,
        expires: datetime | str| int | None = None,
        path: str = "/",
        domain: str | None = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: Literal["lax", "strict", "none"] | None = "lax",
    ) -> None:
        self.response.set_cookie(
            key=key,
            value=value,
            max_age=max_age,
            expires=expires,
            path=path,
            domain=domain,
            secure=secure,
            httponly=httponly,
            samesite=samesite
        )

    def delete_cookie(self, key: str, path: str = "/") -> None:
        return self.set_cookie(key, expires=0, max_age=0, path=path)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from headless.ext import oauth2

from cbra.core.iam.types import Subject
from .authorizationcodecallbackendpoint import AuthorizationCodeCallbackEndpoint


class OIDCRegistrationEndpoint(AuthorizationCodeCallbackEndpoint):
    """Registers a new user or authenticates an existing user with
    a redirect response from an OpenID Connect authorization
    server.
    """
    __module__: str = 'cbra.ext.oauth2'
    summary: str = 'OIDC User Registration Endpoint'

    async def on_success(
        self,
        client: oauth2.Client,
        access_token: oauth2.BearerTokenCredential,
        id_token: oauth2.OIDCToken | None = None
    ) -> None:
        if id_token is None:
            raise NotImplementedError
        subject, onboarded = await self.onboard.oidc(id_token)
        if onboarded:
            await self.on_registered(subject, id_token)
        else:
            await self.on_returning(subject, id_token)

    async def on_registered(
        self,
        subject: Subject,
        token: oauth2.OIDCToken
    ):
        """Invoked when the subject described by the OIDC ID Token
        is a new user.
        """
        await self.session
        self.session.update({
            **token.dict(exclude_none=True),
            'iss': self.get_issuer(),
            'sub': str(subject.uid),
        })

    async def on_returning(
        self,
        subject: Subject,
        token: oauth2.OIDCToken
    ):
        """Invoked when the subject described by the OIDC ID Token
        is a returning user.
        """
        await self.session
        self.session.update({
            **token.dict(include={'email', 'email_verified'}, exclude_none=True),
            'iss': self.get_issuer(),
            'sub': str(subject.uid),
        })


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from types import NotImplementedType
from typing import Any

import pydantic
from ckms.jose import Decoder
from ckms.types import Malformed


class JSONWebTokenPrincipal(pydantic.BaseModel):
    token: str

    @classmethod
    def parse_jwt(
        cls,
        token: str,
        accept: set[str] | NotImplementedType = NotImplemented
    ) -> dict[str, Any]:
        typ = None
        try:
            jose, jwt = Decoder.introspect(token)
        except Malformed:
            raise ValueError('malformed JSON Web Token')
        for header in jose.headers:
            if header.typ is None:
                continue
            typ = str.lower(header.typ)
            break
        else:
            typ = None
        if jwt is None:
            raise ValueError('could not decode JSON Web Token.')
        if typ is None:
            raise ValueError('untyped JWT can not be used as an access token.')
        if accept != NotImplemented and typ not in accept:
            raise TypeError(f'Invalid JWT type: {typ[:16]}')
        return {**jwt.dict(), 'token': token}


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Awaitable
from typing import Callable
from typing import TypeVar

import fastapi


T = TypeVar('T', bound='IDependant')


class IDependant:
    __module__: str = 'cbra.types'

    @classmethod
    def depends(cls: type[T]) -> T:
        """Return a :class:`fastapi.params.Depends` object."""
        return fastapi.Depends(cls.__inject__())

    @classmethod
    def __inject__(cls: type[T]) -> Callable[..., Awaitable[T] | T]:
        """Return a callable that specifies the dependencies of this
        class in its signature.
        """
        return cls


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import uuid

import fastapi
import uvicorn

import cbra


class BookEndpoint(cbra.Endpoint):

    async def get(self, book_id: uuid.UUID):
        print(self.response)


app = fastapi.FastAPI(docs_url='/ui')
BookEndpoint.add_to_router(app, path='/books/{book_id}')


if __name__ == '__main__':
    uvicorn.run('__main__:app', reload=True) # type: ignore



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import base64
from typing import Any

import fastapi
import pydantic

from .icredential import ICredential
from .irequestprincipal import IRequestPrincipal
from .sessionclaims import SessionClaims
from .sessionmodel import SessionModel
from .subjectidentifier import SubjectIdentifier


class SessionRequestPrincipal(IRequestPrincipal, ICredential, SessionModel):
    claims: SessionClaims

    @property
    def subject(self) -> SubjectIdentifier | None:
        if not self.claims.iss or not self.claims.sub:
            return None
        return SubjectIdentifier(
            iss=self.claims.iss,
            sub=self.claims.sub
        )

    @pydantic.root_validator(pre=True)
    def preprocess(
        cls,
        values: dict[str, fastapi.Request]
    ) -> dict[str, Any]:
        if 'request' in values:
            # TODO: This is ugly but we will figure that out later.
            from cbra.core.conf import settings

            cookie = values['request'].cookies.get(settings.SESSION_COOKIE_NAME)
            if cookie is None:
                raise ValueError('session not provided.')
            try:
                data = base64.urlsafe_b64decode(str.encode(cookie, 'ascii'))
            except (ValueError, TypeError):
                raise
            values = cls.parse_raw(data).dict()
        return values

    def has_audience(self) -> bool:
        return False

    def get_credential(self) -> ICredential | None:
        return None


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic
from fastapi.security.utils import get_authorization_scheme_param

from cbra.types import IRequestPrincipal
from ..conf import settings


class SessionRequestPrincipal(IRequestPrincipal, pydantic.BaseModel):
    """A :class:`IRequestPrincipal` that is provided through a cookie.
    """
    @pydantic.root_validator(pre=True)
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        raise Exception
        return cls.parse_authorization_header(values)

    @classmethod
    def parse_authorization_header(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        headers = values.get('headers') or {}
        if not headers.get('Authorization'):
            raise ValueError('no Authorization header.')
        scheme, param = get_authorization_scheme_param(headers['Authorization'])
        if not scheme or not param:
            raise ValueError('the Authorization header was malformed.')
        values.update(cls.parse_scheme(values, str.lower(scheme), param))
        return values

    @classmethod
    def parse_scheme(
        cls,
        values: dict[str, Any],
        scheme: str,
        value: str
    ) -> dict[str, Any]:
        raise NotImplementedError
        



# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi

from cbra.types import RequestPrincipal
from cbra.types import IRoutable
from ..iam import AuthorizationContextFactory
from ..sessions import RequestSession
from .iresource import IResource
from .resourcemodel import ResourceModel
from .resourcetype import ResourceType


class Resource(IResource, metaclass=ResourceType):
    __abstract__: bool = True
    __actions__: list[type[IRoutable]] = []
    __module__: str = 'cbra.core'
    context_factory: AuthorizationContextFactory = AuthorizationContextFactory.depends()
    model: type[ResourceModel]
    principal: RequestPrincipal = RequestPrincipal.depends()
    session: RequestSession = RequestSession.depends()

    def __init_subclass__(cls, model: type[ResourceModel]):
        cls.model = model

    @classmethod
    def add_to_router(cls, router: fastapi.FastAPI, **kwargs: Any) -> None:
        for action in cls.__actions__:
            action.add_to_router(cls, router, **kwargs)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .idependant import IDependant
from .irequestprincipal import IRequestPrincipal
from .isubject import ISubject


class ISubjectResolver(IDependant):
    """Resolves an :class:`IRequestPrincipal` to a :class:`ISubject`
    implementation.
    """
    __module__: str = 'cbra.types'

    async def resolve(self, principal: IRequestPrincipal) -> ISubject:
        raise NotImplementedError


# Copyright (C) 2022 Cochise Ruhulessin # type: ignore
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest
import pytest_asyncio
from headless.core import httpx

import cbra.core as cbra
from cbra import types
from cbra.ext.picqer import PicqerWebhookEndpoint
from cbra.ext.picqer import WebhookSecret

VALID_SIGNATURE: str = '3TMYo2hAFdgQ9bF4MJZ4/dnELdJv5ucJPRAEnoE3Qxk='

VALID_BODY: str = (
    '{"idhook":14381,"name":"created.picklists.picqer.localhost:8000","event":'
    '"picklists.created","event_triggered_at":"2023-02-25 04:34:57","data":{"i'
    'dpicklist":81146972,"picklistid":"P2023-1289","idcustomer":40454718,"idor'
    'der":118330126,"idreturn":null,"idwarehouse":6790,"idtemplate":6375,"idsh'
    'ippingprovider_profile":null,"deliveryname":"ACME","deliverycontact":"Wil'
    'liam Jobs","deliveryaddress":"Renwick Drive 70","deliveryaddress2":"","de'
    'liveryzipcode":"19108","deliverycity":"Philadelphia","deliveryregion":"PA'
    '","deliverycountry":"US","telephone":null,"emailaddress":"n.defeber@molan'
    'o.nl","reference":null,"assigned_to_iduser":null,"invoiced":false,"urgent'
    '":false,"preferred_delivery_date":null,"status":"new","totalproducts":10,'
    '"totalpicked":0,"snoozed_until":null,"closed_by_iduser":null,"closed_at":'
    'null,"created":"2023-02-25 04:34:56","updated":"2023-02-25 04:34:57","pro'
    'ducts":[{"idpicklist_product":190774372,"idproduct":23286044,"idorder_pro'
    'duct":295834751,"idreturn_product_replacement":null,"idvatgroup":14586,"p'
    'roductcode":"XXXBAZ","name":"Test Product Baz","remarks":"","amount":5,"a'
    'mountpicked":0,"amount_picked":0,"price":1,"weight":0,"stocklocation":nul'
    'l,"stock_location":null,"partof_idpicklist_product":null,"has_parts":fals'
    'e},{"idpicklist_product":190774373,"idproduct":23286043,"idorder_product"'
    ':295834752,"idreturn_product_replacement":null,"idvatgroup":14586,"produc'
    'tcode":"XXXBAR","name":"Test Product Bar","remarks":"","amount":5,"amount'
    'picked":0,"amount_picked":0,"price":1,"weight":0,"stocklocation":null,"st'
    'ock_location":null,"partof_idpicklist_product":null,"has_parts":false}],"'
    'comment_count":0}}'
)



@pytest.fixture
def app() -> cbra.Application:
    return cbra.Application()


@pytest_asyncio.fixture # type: ignore
async def client(app: cbra.Application):
    class T(PicqerWebhookEndpoint):
        async def get_webhook_secret(self) -> types.Verifier:
            return WebhookSecret('test')

    app.add(T, path='/')
    async with httpx.Client(base_url='http://cbra.ext.google', app=app) as client:
        yield client


@pytest.mark.asyncio
async def test_no_signature_is_refused(
    client: httpx.Client
):
    response = await client.post(
        url='/',
        content=VALID_BODY,
        headers={'Content-Type': "application/json"}
    )
    assert response.status_code == 403


@pytest.mark.asyncio
async def test_invalid_signature_is_refused(
    client: httpx.Client
):
    response = await client.post(
        url='/',
        content=VALID_BODY,
        headers={
            'Content-Type': "application/json",
            'X-Picqer-Signature': 'aa'
        }
    )
    assert response.status_code == 403


@pytest.mark.asyncio
async def test_valid_signature_is_accepted(
    client: httpx.Client
):
    response = await client.post(
        url='/',
        content=VALID_BODY,
        headers={
            'Content-Type': "application/json",
            'X-Picqer-Signature': VALID_SIGNATURE
        }
    )
    assert response.status_code == 200


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from .icredentialverifier import ICredentialVerifier
from .isubject import ISubject


class NullSubject(ISubject):
    """An :class:`ISubject` implementation that represents an non-authenticated
    and non-identified subject.
    """
    __module__: str = 'cbra.types'
    email: None = None
    sub: None = None

    def is_authenticated(self) -> bool:
        return False

    async def authenticate(
        self,
        verifier: ICredentialVerifier[Any],
        providers: set[str] | None = None
    ) -> None:
        return

    def has_claim(self, name: str, value: Any) -> bool:
        return False


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic


class ResourceListMetadata(pydantic.BaseModel):
    pagination_token: str = pydantic.Field(
        default="",
        alias='paginationToken',
        title='Pagination token',
        description=(
            'This field represents the pagination token to '
            'retrieve the next page of results. If the value '
            'is "", it means no further results for the request.'
        )
    )


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import inspect
from typing import Any

import fastapi.params

from .requirement import Requirement


class Injected(Requirement, fastapi.params.Depends):
    _is_coroutine = asyncio.coroutines._is_coroutine # type: ignore

    @property
    def __signature__(self) -> inspect.Signature:
        return inspect.Signature()

    def __init__(self, name: str, missing: Any | None = NotImplemented):
        Requirement.__init__(self, name=name, missing=missing)
        fastapi.params.Depends.__init__(
            self,
            dependency=self,
            use_cache=False
        )

    async def resolve(self, *args: Any, **kwargs: Any) -> Any:
        assert self.ref is not None
        return self.ref.symbol

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return await self.resolve(*args, **kwargs)


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from cbra.types import IEndpoint
from .persister import Persister
from .resourcemodel import ResourceModel


class IResource(IEndpoint, Persister):
    model: type[ResourceModel]
    path_name: str
    resource_name: str
    resource_id: Any
    verbose_name: str
    verbose_name_plural: str

    async def get_object(self) -> Any:
        raise NotImplementedError


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from datetime import timezone
from typing import Any
from typing import TypeVar

import fastapi

from cbra.types import IAuthorizationContext
from cbra.types import IEndpoint
from cbra.types import RequestPrincipal
from .iam import AuthorizationContextFactory
from .endpointtype import EndpointType
from .sessions import RequestSession


T = TypeVar('T')


class Endpoint(IEndpoint, metaclass=EndpointType):
    __abstract__: bool = True
    __module__: str = 'cbra'
    allowed_http_methods: list[str]
    include_in_schema: bool = True
    principal: RequestPrincipal = RequestPrincipal.depends()
    ctx: IAuthorizationContext
    context_factory: AuthorizationContextFactory = AuthorizationContextFactory.depends()
    session: RequestSession = RequestSession.depends()
    timestamp: datetime

    @classmethod
    def configure(
        cls: type[T],
        overrides: dict[str, Any],
    ) -> type[T]:
        return type(cls.__name__, (cls,), overrides) # type: ignore

    def __init__(self, **kwargs: Any):
        """Constructor. Called in the router; can contain helpful extra
        keyword arguments, and other things.
        """
        # Go through keyword arguments, and either save their values to our
        # instance, or raise an error.
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.timestamp = datetime.now(timezone.utc)

    @classmethod
    def add_to_router(cls, router: fastapi.FastAPI, **kwargs: Any) -> None:
        kwargs.setdefault('path', '/')
        kwargs.setdefault('status_code', cls.status_code)
        kwargs.setdefault('summary', cls.summary)
        kwargs.setdefault('tags', cls.tags)
        for handler in cls.handlers:
            handler.add_to_router(cls, router, **kwargs)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pytest
import pytest_asyncio
from headless.core import httpx

import cbra.core as cbra
from cbra.ext.google import GoogleEndpoint
from cbra.types import JSONWebTokenPrincipal


class GoogleServiceEndpoint(GoogleEndpoint):

    async def get(self) -> dict[str, Any]:
        return self.principal.dict()


@pytest.fixture
def app() -> cbra.Application:
    return cbra.Application()


@pytest_asyncio.fixture # type: ignore
async def client(app: cbra.Application):
    async with httpx.Client(base_url='http://cbra.ext.google', app=app) as client:
        yield client


@pytest.mark.asyncio
async def test_basic_oidc_authentication_is_refused_without_token(
    app: cbra.Application,
    client: httpx.Client,
    google_id_token: str
):
    app.add(GoogleServiceEndpoint)
    response = await client.get(url='/')
    assert response.status_code == 401
    pass


@pytest.mark.asyncio
async def test_basic_oidc_authentication(
    app: cbra.Application,
    client: httpx.Client,
    google_id_token: str
): 
    claims: dict[str, Any] = JSONWebTokenPrincipal.parse_jwt(google_id_token)
    app.add(GoogleServiceEndpoint.configure(allowed_service_accounts={claims['email']}))
    response = await client.get(
        url='/',
        headers={'Authorization': f'Bearer {google_id_token}'}
    )
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_basic_oidc_authentication_accepts_only_whitelisted(
    app: cbra.Application,
    client: httpx.Client,
    google_id_token: str
): 
    app.add(GoogleServiceEndpoint.configure(allowed_service_accounts=set()))
    response = await client.get(
        url='/',
        headers={'Authorization': f'Bearer {google_id_token}'}
    )
    assert response.status_code == 403


@pytest.mark.asyncio
async def test_basic_oidc_authentication_accepts_only_google(
    app: cbra.Application,
    client: httpx.Client,
    google_id_token: str
):
    pytest.skip("TODO")


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

import pydantic
from headless.ext.oauth2.models import SubjectIdentifier

from .principalmodel import PrincipalModel


class ExternalPrincipal(PrincipalModel):
    """Represents a subject identity that was asserted
    by an external identity provider.
    """
    kind: Literal['SubjectIdentifier']
    iss: str = pydantic.Field(..., primary_key=True)
    sub: str = pydantic.Field(..., primary_key=True)

    @pydantic.root_validator(pre=True)
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        value: SubjectIdentifier | None = values.pop('principal', None)
        if isinstance(value, SubjectIdentifier):
            values.update({
                'iss': value.iss,
                'sub': value.sub
            })
        return values


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import hashlib
from typing import Any

from canonical import EmailAddress
from headless.ext.oauth2.models import SubjectIdentifier

from .types import PublicIdentifier
from .types import PrincipalType


class PrincipalHasher:
    """Known how to create a unique hash of a given principal."""
    __module__: str = 'cbra.core.iam'

    #: The algorithm that is used to hash principals.
    #:
    #: .. warning::
    #:
    #:   Changing this for an existing dataset will cause duplicate
    #:   principals.
    algorithm: str = 'sha1'

    def create_hasher(self) -> Any:
        # We use SHA-1 here because its unique enough, as it will
        # be applied on data that is already being constrained
        # such as email addresses or phone numbers. If there is
        # a collision then the user just has bad luck.
        return hashlib.new(self.algorithm)

    @functools.singledispatchmethod
    def hash(
        self,
        principal: PrincipalType,
    ) -> str:
        """Create a hash of the given principal."""
        raise NotImplementedError(type(principal).__name__)

    @hash.register # type: ignore
    def hash_email(
        self,
        principal: EmailAddress
    ) -> str:
        h = self.create_hasher()
        h.update(str.encode(type(principal).__name__, 'ascii'))
        h.update(str.encode(principal))
        return self._format(h.hexdigest())

    @hash.register # type: ignore
    def hash_subject_identifier(
        self,
        principal: SubjectIdentifier
    ) -> str:
        h = self.create_hasher()
        h.update(str.encode(type(principal).__name__, 'ascii'))
        h.update(str.encode(principal.iss))
        h.update(str.encode(principal.sub))
        return self._format(h.hexdigest())

    @hash.register # type: ignore
    def hash_public_identifier(
        self,
        principal: PublicIdentifier
    ) -> str:
        h = self.create_hasher()
        h.update(str.encode(type(principal).__name__, 'ascii'))
        h.update(str.encode(principal.iss))
        h.update(str.encode(principal.sub))
        return self._format(h.hexdigest())
    
    def _format(self, v: str) -> str:
        return f'{self.algorithm}:{v}'


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Generic
from typing import TypeVar

from .ihashable import IHashable
from .sessionclaims import SessionClaims


C = TypeVar('C', bound=SessionClaims)
T = TypeVar('T', bound='ISessionData[Any]')


class ISessionData(Generic[C], IHashable):
    claims: C | None

    @classmethod
    def parse_cookie(cls: type[T], value: str) -> T | None:
        raise NotImplementedError

    def as_cookie(self) -> str:
        raise NotImplementedError

    def get(self, key: str) -> Any:
        raise NotImplementedError

    def set(self, key: str, value: Any) -> bool:
        """Set the value and return a boolean indicating if the data
        was modified.
        """
        raise NotImplementedError

    def update(self, claims: dict[str, Any]) -> None:
        raise NotImplementedError

    async def sign(
        self,
        sign: Callable[[bytes], Awaitable[str]]
    ) -> None:
        raise NotImplementedError

    async def verify(
        self,
        verify: Callable[[bytes | str, bytes], Awaitable[bool]]
    ) -> bool:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import runtime_checkable
from typing import Any
from typing import AsyncGenerator
from typing import Protocol

from .principal import Principal
from .principaltype import PrincipalType
from .subject import Subject


@runtime_checkable
class ISubjectRepository(Protocol):
    __module__: str = 'cbra.core.iam'

    async def find_by_principals(self, principals: list[Any]) -> set[int]: ...
    async def get(self, principal: int | PrincipalType) -> Subject: ...
    async def persist(self, subject: Subject) -> Subject: ...
    def get_principals(self, subject_id: int) -> AsyncGenerator[Principal, None]: ...


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import uuid
from typing import Any

from .pathparameter import PathParameter
from .notfound import NotFound


class UUIDPathParameter(PathParameter):
    __module__: str = 'cbra.types'
    format: str = 'uuid'
    type: str = 'string'

    @classmethod
    def clean(cls, v: Any) -> uuid.UUID:
        try:
            return uuid.UUID(v)
        except (TypeError, ValueError):
            raise NotFound


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import overload
from typing import TypeVar

import pydantic
import pydantic.main


T = TypeVar('T', bound='ModelAutoAssignedIdentity')


class ModelAutoAssignedIdentity:
    """Like :class:`~cbra.types.ModelIdentity`, but for a single field
    that is an auto assigned integer.
    """
    __module__: str = 'cbra.types'
    field: pydantic.main.FieldInfo
    attname: str

    def __init__(self, attname: str, field: pydantic.main.FieldInfo):
        self.attname = attname
        self.field = field

    @overload
    def __get__(
        self: T,
        obj: None,
        cls: None
    ) -> T:
        ...

    @overload
    def __get__(
        self,
        obj: pydantic.BaseModel,
        cls: type[pydantic.BaseModel]
    ) -> int:
        ...

    def __get__(
        self: T,
        obj: pydantic.BaseModel | None,
        cls: type[pydantic.BaseModel] | None
    ) -> int | None | T:
        if obj is None:
            return self
        return getattr(obj, self.attname)

    def __set__(self, obj: pydantic.BaseModel, value: int) -> None:
        setattr(obj, self.attname, value)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol
from typing import TypeVar

from .iauthorizationflowstate import IAuthorizationFlowState

M = TypeVar('M')


class IAuthorizationServerStorage(Protocol):
    __module__: str = 'cbra.ext.oauth2.types'


    async def persist(
        self,
        obj: Any
    ) -> None:
        ...

    async def destroy(self, obj: Any) -> None: ...
    async def get_state(self, key: str) -> IAuthorizationFlowState | None:
        ...


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

from cbra.types import Operation
from .detailaction import DetailAction


T = TypeVar('T', bound='DeleteAction')


class DeleteAction(DetailAction):
    action: str = 'destroy'
    name_template: str = 'Delete {article} {name}'
    response_description: str = 'The deleted **{name}**'

    @classmethod
    def fromfunc(
        cls: type[T],
        name: str,
        func: Any
    ) -> T:
        return cls(name=name, method='DELETE', func=func)

    def can_write(self) -> bool:
        return False

    def get_openapi_responses(
        self,
        responses: dict[int | str, Any]
    ) -> dict[int | str, Any]:
        responses.update({
            202: {
                'description': (
                    f'The **{self.endpoint.verbose_name}** is scheduled for '
                    'deletion'
                ),
                'model': Operation
            }
        })
        return super().get_openapi_responses(responses)

    def get_return_annotation(self) -> Any:
        return self.endpoint.model.__response_model__ | Operation


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

from canonical import EmailAddress
from headless.ext.oauth2.models import OIDCToken

from .subject import Subject
from .principaltype import PrincipalType


class IUserOnboardingService(Protocol):
    def initialize(self) -> Subject: ...
    async def can_use(self, subject: Subject, principals: list[PrincipalType]) -> bool: ...
    async def email(self, issuer: str, email: EmailAddress) -> tuple[Subject, bool]: ...
    async def oidc(self, token: OIDCToken) -> tuple[Subject, bool]: ...
    async def sync(self, issuer: str, principal: PrincipalType) -> tuple[Subject, bool]: ...
    async def update(self, subject: Subject) -> None: ...


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from cbra.core import Application
from cbra.core.utils import parent_signature
from .aortaendpoint import AortaEndpoint
from .environ import GOOGLE_DATASTORE_NAMESPACE
from .environ import GOOGLE_SERVICE_PROJECT


class Service(Application):
    __module__: str = 'cbra.ext.google'

    @parent_signature(Application.__init__)
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        if GOOGLE_SERVICE_PROJECT and GOOGLE_DATASTORE_NAMESPACE:
            from google.cloud.datastore import Client
            self.inject(
                name='GoogleDatastoreClient',
                value=Client(
                    project=GOOGLE_SERVICE_PROJECT,
                    namespace=GOOGLE_DATASTORE_NAMESPACE
                )
            )
            self.container.provide('SubjectResolver', {
                'qualname': 'cbra.ext.google.DatastoreSubjectResolver'
            })
            self.container.provide('SubjectRepository', {
                'qualname': 'cbra.ext.google.DatastoreSubjectRepository'
            })
        self.add(AortaEndpoint, path="/.well-known/aorta")


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cbra.ext.google import DatastoreRepository
from cbra.ext.oauth2 import BaseStorage
from cbra.ext.oauth2.models import AuthorizationState
from cbra.types import PersistedModel


class Storage(BaseStorage, DatastoreRepository):
    __module__: str = 'cbra.ext.google.impl.oauth2'

    async def destroy(self, obj: PersistedModel) -> None:
        return await self.delete(self.model_key(obj))

    async def get_state(self, key: str) -> AuthorizationState | None:
        return await self.get_model_by_key(AuthorizationState, key)

    async def persist_state(self, obj: AuthorizationState) -> None:
        await self.put(obj)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import TypeVar
from typing import Union

import fastapi
import pydantic
from headless.ext.oauth2.types import ResponseMode
from headless.ext.oauth2.types import ResponseType

from cbra.types import IDependant
from cbra.core.ioc import override
from .authorizationrequestobject import AuthorizationRequestObject
from .authorizationrequestreference import AuthorizationRequestReference
from .redirecturi import RedirectURI


T = TypeVar('T', bound='BaseAuthorizationRequest')


class BaseAuthorizationRequest(IDependant, pydantic.BaseModel):

    @classmethod
    def fromrequest(
        cls: type[T],
        client_id: str | None = fastapi.Query(
            default=None,
            title="Client ID",
            description="Identifies the client that is requesting authorization."
        ),
        response_type: ResponseType | None = fastapi.Query(
            default=None,
            title="Response type",
            description=(
                "Informs the authorization server of the desired response type."
            ),
            example="code",
        ),
        redirect_uri: RedirectURI | None = fastapi.Query(
            default=None,
            title="Redirect URI",
            description=(
                "The URL to redirect the client to after completing the "
                "flow. Must be an absolute URI that is served over https.\n\n"
                "If `redirect_uri` is omitted, the default redirect URI for "
                "the client specified by `client_id` is used. For clients that "
                "do not have a redirect URI specified, this produces an error "
                "state."
            )
        ),
        scope: str | None = fastapi.Query(
            default=None,
            title="Scope",
            description=(
                "A space-delimited list specifying the requested access scope."
            ),
            max_length=512,
            example="hello.world"
        ),
        state: str | None = fastapi.Query(
            default=None,
            title="State",
            description=(
                "An opaque value used by the client to maintain state between "
                "the request and callback. The authorization server includes "
                "this value when redirecting the user-agent back to the client."
            ),
            max_length=64,
            example=bytes.hex(os.urandom(64))
        ),
        response_mode: ResponseMode | None = fastapi.Query(
            default=None,
            title="Response mode",
            description=(
                "Informs the authorization server of the mechanism to be used "
                "for returning authorization response parameters."
            ),
            example="query"
        ),
        request: str | None = fastapi.Query(
            default=None,
            title="Request",
            description=(
                "A JSON Web Token (JWT) whose JWT Claims Set holds the "
                "JSON-encoded OAuth 2.0 authorization request parameters. "
                "Must not be used in combination with the `request_uri` "
                "parameter, and all other parameters except `client_id` "
                "must be absent.\n\n"
                "Confidential and credentialed clients must first sign "
                "the claims using their private key, and then encrypt the "
                "result with the public keys that are provided by the "
                "authorization server through the `jwks_uri` specified "
                "in its metadata."
            )
        ),
        request_uri: str | None = fastapi.Query(
            default=None,
            title="Request URI",
            description=(
                "References a Pushed Authorization Request (PAR) or a remote "
                "object containing the authorization request.\n\n"
                "If the authorization request was pushed to this authorization "
                "server, then the format of the `request_uri` parameter is "
                "`urn:ietf:params:oauth:request_uri:<reference-value>`. "
                "Otherwise, it is an URI using the `https` scheme. If the "
                "`request_uri` parameter is a remote object, then the external "
                "domain must have been priorly whitelisted by the client."
            )
        )
    ) -> T:
        raise NotImplementedError

    @classmethod
    def __inject__(cls: type[T]) -> Callable[..., Awaitable[T] | T]:
        return cls.fromrequest


class AuthorizationRequest(BaseAuthorizationRequest):
    __root__: Union[
        AuthorizationRequestReference,
        AuthorizationRequestObject
    ]

    @classmethod
    @override(BaseAuthorizationRequest.fromrequest) # type: ignore
    def fromrequest(cls: type[T], **kwargs: Any) -> T: # type: ignore
        return cls.parse_obj(kwargs)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from .icredential import ICredential
from .httpheaderprincipal import HTTPHeaderPrincipal
from .jsonwebtokenprincipal import JSONWebTokenPrincipal
from .jsonwebtoken import JSONWebToken


class OIDCRequestPrincipal(HTTPHeaderPrincipal, JSONWebTokenPrincipal):
    iss: str
    sub: str
    exp: int
    aud: str| list[str]
    iat: int
    auth_time: int | None = None
    nonce: str | None = None
    acr: str = "0"
    amr: list[str] = []
    azp: str | None = None
    token: str

    # Standard claims
    email: str | None = None

    @classmethod
    def parse_scheme(
        cls,
        values: dict[str, Any],
        scheme: str,
        value: str
    ) -> dict[str, Any]:
        if scheme != 'bearer':
            raise ValueError('this principal requires the Bearer scheme')
        values.update(cls.parse_jwt(value, accept={"application/jwt", "jwt"}))
        return values

    def get_audience(self) -> set[str]:
        audience = self.aud
        if isinstance(audience, str):
            audience = {audience}
        return set(audience)

    def get_credential(self) -> ICredential | None:
        return JSONWebToken(self.token)

    def has_audience(self) -> bool:
        return True

    def validate_audience(self, allow: set[str]) -> bool:
        aud = self.aud
        if isinstance(aud, str):
            aud = {aud}
        return bool(set(aud) & allow)


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import TypeVar


_T = TypeVar('_T')


def parent_signature(f: _T) -> Callable[[Any], _T]:
    def decorator_factory(func: Callable[[Any], _T]) -> _T:
        sig = inspect.signature(f) # type: ignore
        func.__signature__ = sig # type: ignore
        return func # type: ignore
    return decorator_factory


async def ensure_awaited(obj: _T | Awaitable[_T], default: Any = NotImplemented) -> _T:
    if inspect.isawaitable(obj):
        value: _T = await obj
    else:
        value = obj # type: ignore
    if not value and (default != NotImplemented):
        value = default
    return value


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
.. _ref-guides-integrating-an-oauth2-authorization-server

=============================================================
Implementing an OAuth 2.x/OpenID Connect authorization server
=============================================================
"""
from .authorizationendpoint import AuthorizationEndpoint
from .authorizationcodecallbackendpoint import AuthorizationCodeCallbackEndpoint
from .basestorage import BaseStorage
from .loginendpoint import LoginEndpoint
from .memorystorage import MemoryStorage
from .oidcregistrationendpoint import OIDCRegistrationEndpoint


__all__: list[str] = [
    'AuthorizationEndpoint',
    'AuthorizationCodeCallbackEndpoint',
    'BaseStorage',
    'LoginEndpoint',
    'MemoryStorage',
    'OIDCRegistrationEndpoint',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

from headless.ext.picqer.v1 import Order

from .event import Event


class OrderEvent(Event):
    event: Literal[
        "orders.created",
        "orders.allocated",
        "orders.closed",
        "orders.status_changed",
        "orders.completed",
    ]

    data: Order



# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .abortable import Abortable
from .basemodel import BaseModel
from .basemodel import BaseModelMetaclass
from .conflict import Conflict
from .forbidden import Forbidden
from .hints import PolymorphicIterable
from .iauthorizationcontext import IAuthorizationContext
from .iauthorizationcontextfactory import IAuthorizationContextFactory
from .icredential import ICredential
from .icredentialverifier import ICredentialVerifier
from .icursor import ICursor
from .ideferred import IDeferred
from .idependant import IDependant
from .iendpoint import IEndpoint
from .ihashable import IHashable
from .imodelrepository import IModelRepository
from .integerpathparameter import IntegerPathParameter
from .iroutable import IRoutable
from .irequestprincipal import IRequestPrincipal
from .irequestprincipalintrospecter import IRequestPrincipalIntrospecter
from .isessionfactory import ISessionFactory
from .isessionmanager import ISessionManager
from .isubject import ISubject
from .isubjectresolver import ISubjectResolver
from .jsonwebtoken import JSONWebToken
from .jsonwebtokenprincipal import JSONWebTokenPrincipal
from .mutablesignature import MutableSignature
from .nullrequestprincipal import NullRequestPrincipal
from .nullsubject import NullSubject
from .nullsubjectesolver import NullSubjectResolver
from .notauthorized import NotAuthorized
from .notfound import NotFound
from .operation import Operation
from .oidcrequestprincipal import OIDCRequestPrincipal
from .pathparameter import PathParameter
from .persistedmodel import PersistedModel
from .requestprincipal import RequestPrincipal
from .rfc9068requestprincipal import RFC9068RequestPrincipal
from .session import Session
from .sessionmodel import SessionModel
from .sessionrequestprincipal import SessionRequestPrincipal
from .stringpathparameter import StringPathParameter
from .subjectidentifier import SubjectIdentifier
from .unauthenticatedauthorizationcontext import UnauthenticatedAuthorizationContext
from .uuidpathparameter import UUIDPathParameter
from .verifier import Verifier


__all__: list[str] = [
    'Abortable',
    'BaseModel',
    'BaseModelMetaclass',
    'Conflict',
    'Forbidden',
    'ICredential',
    'ICredentialVerifier',
    'ICursor',
    'IDependant',
    'IDeferred',
    'IAuthorizationContext',
    'IAuthorizationContextFactory',
    'IEndpoint',
    'IModelRepository',
    'IntegerPathParameter',
    'IHashable',
    'IRequestPrincipalIntrospecter',
    'IRequestPrincipal',
    'IRoutable',
    'ISessionFactory',
    'ISessionManager',
    'ISubject',
    'ISubjectResolver',
    'JSONWebToken',
    'JSONWebTokenPrincipal',
    'MutableSignature',
    'NotAuthorized',
    'NotFound',
    'NullRequestPrincipal',
    'NullSubject',
    'NullSubjectResolver',
    'Operation',
    'OIDCRequestPrincipal',
    'PathParameter',
    'PersistedModel',
    'PolymorphicIterable',
    'RequestPrincipal',
    'RFC9068RequestPrincipal',
    'Session',
    'SessionModel',
    'SessionRequestPrincipal',
    'SubjectIdentifier',
    'StringPathParameter',
    'UnauthenticatedAuthorizationContext',
    'UUIDPathParameter',
    'Verifier',
]


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
""".. _guides-authentication:

=======================
Authenticating requests
=======================
The :mod:`cbra` module comes with a basic authentication system that enables
resolving an incoming HTTP request to a subject identifier, based on the
parameters included in the request, such as headers or cookies. A subject
identifier uniquely points to an account of some kind (user account, service
account) that has access to certain protected resources.

What is authentication?
=======================
Authentication is a term that refers to the process of proving that some fact
or some document is genuine. In computer science, this term is typically
associated with proving a user's identity. The process of authentication can
thus be modelled into three models:

- A **Subject** is a person or a service that has one or many identities.
- A **Principal** represents a specific identity of a Subject, such as
  an email address, phone number or account identifier.
- A **Credential** is used to assert the ownership of or access to a
  specific Principal.

Practically speaking, *authentication* may then be defined as follows:

A **Subject** uses a combination of a **Principal** and a **Credential** to
assert its identity. These concepts are reflected in the abstract interfaces
specified by :mod:`cbra.types`, being :class:`~cbra.types.ICredential`,
:class:`~cbra.types.IRequestPrincipal` and :class:`~cbra.types.ISubject`. The default
implementation in :mod:`cbra.core.iam` supports RFC 9068 and
OpenID Connect ID Tokens to authenticate requests, but can quite easily
extended to support other mechanisms, like your own bearer token.


Configuring authentication
==========================
Configuring :mod:`cbra` with the default implementation is as simple as setting
:attr:`~cbra.core.conf.settings.TRUSTED_AUTHORIZATION_SERVERS` in your settings
module:

.. code:: python

    # settings.py
    TRUSTED_AUTHORIZATION_SERVERS: list[str] = [
        'https://accounts.google.com'
    ]


Authenticated requests
======================
The :mod:`cbra` IAM framework may be used with :class:`~cbra.core.Endpoint`
and :class:`cbra.core.Resource`, or plain functions as you might be used to
with FastAPI:

.. code:: python

    import fastapi
    import cbra.core as cbra
    from cbra.core.iam import AuthorizationContextFactory
    from cbra.types import Principal

    app: cbra.Application = cbra.Application()

    @app.get('/')
    async def f(
        request: fastapi.Request,
        factory: AuthorizationContextFactory = fastapi.Depends(),
        principal: Principal = fastapi.Depends(Principal.fromrequest)
    ):
        ctx = await factory.authenticate(request, principal)
        return {
            'remote_host': ctx.remote_host,
            'is_authenticated': ctx.is_authenticated()
        }

    if __name__ == '__main__':
        import uvicorn
        uvicorn.run(app)

Similarly, :class:`~cbra.core.Endpoint` and :class:`~cbra.core.Resource`
may be used with the authentication framework:

.. code:: python

    import cbra.core as cbra


    class AuthenticatedEndpoint(cbra.Endpoint):

        async def get(self) -> dict[str, bool | str]:
            return {
                'remote_host': str(self.ctx.remote_host),
                'is_authenticated': self.ctx.is_authenticated()
            }


    app: cbra.Application = cbra.Application()
    app.inject('TRUSTED_AUTHORIZATION_SERVERS', ['https://accounts.google.com'])
    app.add(AuthenticatedEndpoint)


    if __name__ == '__main__':
        import uvicorn
        uvicorn.run('__main__:app', reload=True)

To run the above example, the following script obtains an OpenID Connect
ID Token from Google:

.. code:: python

    import google.oauth2.id_token
    import google.auth.transport.requests

    if __name__ == '__main__':
        request = google.auth.transport.requests.Request()
        audience = 'http://localhost:8000'
        id_token = google.oauth2.id_token.fetch_id_token(request, audience)
        headers = {'Authorization': f'bearer {id_token}'}
        print(id_token)
"""
from .authenticatedcontext import AuthenticatedContext
from .authenticationservice import AuthenticationService
from .authorizationcontextfactory import AuthorizationContextFactory
from .isubjectrepository import ISubjectRepository
from .subjectresolver import SubjectResolver
from .types import IUserOnboardingService


__all__: list[str] = [
    'AuthenticatedContext',
    'AuthenticationService',
    'AuthorizationContextFactory',
    'ISubjectRepository',
    'IUserOnboardingService',
    'SubjectResolver',
]